package com.sf.vertx.init;

import java.util.List;
import java.util.concurrent.ConcurrentHashMap;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.ApplicationArguments;
import org.springframework.boot.ApplicationRunner;
import org.springframework.core.annotation.Order;
import org.springframework.stereotype.Component;

import com.sf.vertx.api.pojo.AppConfig;
import com.sf.vertx.api.pojo.Node;
import com.sf.vertx.api.pojo.VertxConfig;
import com.sf.vertx.arithmetic.roundRobin.SacLoadBalancing;
import com.sf.vertx.arithmetic.roundRobin.WeightedRoundRobin;
import com.sf.vertx.constans.RedisConfig;
import com.sf.vertx.service.AppConfigService;

import io.vertx.core.Future;
import io.vertx.core.Vertx;
import io.vertx.core.VertxOptions;
import io.vertx.core.http.HttpClient;
import io.vertx.core.http.HttpServer;
import io.vertx.core.http.HttpServerRequest;
import io.vertx.core.net.SocketAddress;
import io.vertx.ext.web.Router;
import io.vertx.ext.web.proxy.handler.ProxyHandler;
import io.vertx.httpproxy.HttpProxy;
import lombok.extern.slf4j.Slf4j;

/***
 * 动态构建vertx服务
 * 
 * @author xy
 *
 */
@Slf4j
@Order(value = 10)
@Component
public class DynamicBuildServer implements ApplicationRunner {
	@Value("${server.vertx.server.default.port}")
    private Integer serverDefaultPort;
	
	@Autowired
	private AppConfigService appConfigService;
	
	@Autowired
	private RedisConfig redisConfig;

	@Override
	public void run(ApplicationArguments args) throws Exception {
		// 初始化redis key
		redisConfig.init();
		
		// 加载vertx、应用配置
		appStartLoadData();
	}

	/***
	 * 应用启动, 从redis读取配置,初始化vertx服务
	 */
	private void appStartLoadData() {
		// 编解码线程池
		Vertx VERTX = Vertx.vertx(new VertxOptions().setWorkerPoolSize(20));
		// 创建HTTP监听
		HttpServer server = VERTX.createHttpServer();
		Router mainHttpRouter = Router.router(VERTX);
		VertxConfig vertxConfig = appConfigService.loadVertxConfig();
		Integer serverPort = (vertxConfig == null || vertxConfig.getPort() == null) ? serverDefaultPort : vertxConfig.getPort();
		server.requestHandler(mainHttpRouter).listen(serverPort, h -> {
			if (h.succeeded()) {
				log.info("HTTP端口监听成功:{}", serverPort);
			} else {
				log.error("HTTP端口监听失败:{}", serverPort);
			}
		});
		
		ConcurrentHashMap<String, AppConfig> cacheAppConfig = appConfigService.loadAllConfig();
		for (String appCode : cacheAppConfig.keySet()) {
			AppConfig appConfig = cacheAppConfig.get(appCode);
			// 负载均衡算法
			// 轮训
			SacLoadBalancing sacLoadBalancing = roundRobin(appConfig.getNodeList());
			
			
			// 有策略
			if (appConfig.getStrategyList() != null && appConfig.getStrategyList().size() > 0) {

			} else {
				// 无策略,直接走反向代理
//				HttpClientOptions clientOptions = new HttpClientOptions();
//				clientOptions.setMaxPoolSize(20); // 最大连接池大小
//				clientOptions.setConnectTimeout(5000); // 连接超时 毫秒
//				clientOptions.setHttp2KeepAliveTimeout(1);
//				clientOptions.setIdleTimeout(1000); // 连接空闲超时 毫秒
//				HttpClient proxyClient = VERTX.createHttpClient(clientOptions);
				
				HttpClient proxyClient = VERTX.createHttpClient();
				HttpProxy proxy = HttpProxy.reverseProxy(proxyClient);
				proxy.originSelector(request -> Future.succeededFuture(resolveOriginAddress(sacLoadBalancing, request)));
				mainHttpRouter.route().handler(ProxyHandler.create(proxy));
			}
		}
	}
	
	public SocketAddress resolveOriginAddress(SacLoadBalancing sacLoadBalancing, HttpServerRequest request) {
		// TODO 区分https、http
		Node node = sacLoadBalancing.selectNode();
		SocketAddress socketAddress = SocketAddress.inetSocketAddress(node.getPort(), node.getIp());
		log.info("负载均衡跳转地址:{}", socketAddress.host() +";"+socketAddress.port());
		return socketAddress;
	}

	
	private SacLoadBalancing roundRobin(List<Node> nodeList) {
		WeightedRoundRobin weightedRoundRobin = new WeightedRoundRobin();
		int weight = 1;
		for(Node node : nodeList) {
			node.setWeight(weight);
			node.setCurrentWeight(weight);
			node.setEffectiveWeight(weight);
			weight++;
			WeightedRoundRobin.totalWeight += node.getEffectiveWeight();
		}
		weightedRoundRobin.init(nodeList);
		return weightedRoundRobin;
	}
}
